import json

import names
import numpy as np
import pandas as pd
from randomtimestamp import randomtimestamp

from teaching_tools.ab_test.params import cohort_params


def load_json(filepath: str) -> dict:
    """Loads JSON file.

    Parameters
    ----------
    filepath : str
        Location of JSON file

    Returns
    -------
    dict

    """
    with open(filepath, "r") as f:
        return json.load(f)


def load_cohort_params(c) -> dict:
    """Helper function for loading cohort parameters.

    Parameters
    ----------
    c : Union[str, dict]
        If ``str``, treated as a filepath. If ``dict``, returns same.

    Returns
    -------
    dict

    """
    if isinstance(c, str):
        return load_json(c)
    elif isinstance(c, dict):
        return c
    else:
        return cohort_params


class StudentGenerator:
    def __init__(
        self, cohort_params=None, days=1, n=None, is_experiment=False
    ):
        self.cohort_params = load_cohort_params(cohort_params)
        self.days = days
        if n is None:
            self.n = self.calculate_n_students()
        else:
            self.n = n
        self.is_experiment = is_experiment

    def calculate_n_students(self) -> int:
        mu = self.cohort_params["daily_traffic"]["mu"]
        sigma = self.cohort_params["daily_traffic"]["sigma"]
        n = sum(np.random.normal(mu, sigma, self.days).astype(int))
        return n

    def generate_bday(self, years):
        days_offset = np.random.randint(low=0, high=350)
        birthdate = pd.Timestamp.now() - pd.DateOffset(
            years=int(years), days=days_offset
        )
        return birthdate.replace(hour=0, minute=0, second=0, microsecond=0)

    def generate_categorical(self, labels, probabilities, size=1):
        labels = np.array(labels)
        probabilities = np.array(probabilities)
        data = np.random.choice(
            a=labels,
            size=size,
            p=probabilities,
        )
        return data

    def generate_email(self, first: str, last: str) -> str:
        domains = [
            "@yahow.com",
            "@gmall.com",
            "@microsift.com",
            "@hotmeal.com",
        ]
        email = (
            first.lower()
            + "."
            + last.lower()
            + str(np.random.randint(1, 100))
            + np.random.choice(domains, 1)[0]
        )
        return email

    def generate_students(self):
        df = pd.DataFrame()

        # Add columns from demographic params
        for k, v in self.cohort_params["demographics"].items():
            df[k] = self.generate_categorical(
                labels=list(v.keys()),
                probabilities=list(v.values()),
                size=self.n,
            )

        # Create birthday from age
        df["birthday"] = df["age"].apply(self.generate_bday)
        df.drop(columns=["age"], inplace=True)

        # Create name
        df["first_name"] = df["gender"].apply(
            lambda x: names.get_first_name(gender=x)
        )
        df["last_name"] = [names.get_last_name() for x in range(self.n)]

        # Create email
        df["email"] = df[["first_name", "last_name"]].apply(
            lambda x: self.generate_email(*x), axis=1
        )

        # Create timestamp
        start = self.cohort_params["dates"]["start"]
        stop = self.cohort_params["dates"]["stop"]
        df["created"] = [
            randomtimestamp(start=start, end=stop) for x in range(self.n)
        ]

        # Reorder columns
        cols = [
            "created",
            "first_name",
            "last_name",
            "email",
            "birthday",
            "gender",
            "education_level",
            "nationality",
            "completed_quiz",
        ]
        if self.is_experiment:
            cols.pop(-1)
        df = df[cols]

        if self.is_experiment:
            response_var = self.cohort_params["experiment"]["response_var"]
            contingency_table = self.cohort_params["experiment"][
                "contingency_table"
            ]
            data = []
            for group, response in contingency_table.items():
                for k, v in response.items():
                    count = round(self.n * v)
                    data += [(group, k)] * count

            df[["group", response_var]] = pd.DataFrame(data)
            df["in_experiment"] = True

        return df


class Cohort:
    def __init__(self, **kwargs):
        self._students = None
        self.generator = StudentGenerator(**kwargs)

    def enroll_students(self, confirm_msg=True):
        self._students = self.generator.generate_students()
        if confirm_msg:
            print(f"Added {len(self._students)} students to cohort.")

    def yield_students(self):
        if self._students is None:
            self.enroll_students(confirm_msg=False)
        for s in self._students.itertuples(index=False, name="Student"):
            yield s._asdict()

    def get_students(self) -> pd.DataFrame:
        """Getter function"""
        if self._students is None:
            raise AttributeError(
                "You have not enrolled any students in this cohort. Use `enroll_students`."
            )
        else:
            return self._students



from typing import Union

from pymongo import MongoClient

from teaching_tools.ab_test.data_model import Cohort
from teaching_tools.ab_test.repository import (
    AbstractRepository,
    MongoRepository,
)


class Experiment:
    """Class to generate experimental student data and insert into repository."""

    def __init__(self, repo: AbstractRepository) -> None:
        self.attach_repository(repo)

    def attach_repository(
        self,
        repository: Union[AbstractRepository, MongoClient],
        repo_name="wqu-abtest",
    ) -> None:
        """Attach new repository to experiment.

        Parameters
        ----------
        repository : AbstractRepository

        repo_name : str, optional
            If ``repository`` is ``MongoClient``, name of database, by default "wqu-abtest".

        Raises
        ------
        Exception
            Failed to attach repository.
        """
        if isinstance(repository, MongoClient):
            self.repo = MongoRepository(repository, repo_name)
        elif isinstance(repository, AbstractRepository):
            self.repo = repository
        else:
            raise Exception("Cannot attach repository.")

    def add_cohort_to_repository(
        self, collection="ds_applicants", overwrite=False, confirm_msg=True
    ) -> None:
        """Add cohort of students to repository.

        Parameters
        ----------
        collection : str, optional
            Name of collection to which cohort will be added, by default "ds_applicants"
        overwrite : bool, optional
            Whether to overwrite collection if already exists, by default False
        confirm_msg : bool, optional
            Whether to print confirmation message, by default True
        """
        self.repo.create_collection(collection, overwrite=overwrite)
        r = self.repo.insert(
            collection,
            records=self.cohort.yield_students(),
            return_result=True,
        )
        if confirm_msg:
            if r["acknowledged"]:
                print(
                    f"Added {len(self.cohort._students)} students to repository."
                )
            else:
                print(
                    f"Failed to add {len(self.cohort._students)} students to repository."
                )

    def run_experiment(self, days: int) -> None:
        """Add experimental student data to repository.

        Parameters
        ----------
        days : int
            Number of days to run the experiment. More days means more students.
        """
        self.cohort = Cohort(days=days, is_experiment=True)
        self.add_cohort_to_repository()

    def reset_experiment(self, confirm_msg=True) -> None:
        """Remove experimental student data from repository.

        Parameters
        ----------
        confirm_msg : bool, optional
            Whether to print confirmation message, by default True
        """
        results = self.repo.drop(
            collection="ds_applicants",
            query={"in_experiment": True},
            return_result=True,
        )
        if results["acknowledged"] and confirm_msg:
            print(f"Deleted {results['deleted_count']} from repository.")



# Inspiration: https://tinyurl.com/2am2r7ga
import abc


class AbstractRepository(abc.ABC):
    @abc.abstractmethod
    def create_collection(self, collection, overwrite=True):
        raise NotImplementedError

    @abc.abstractmethod
    def drop_collection(self, collection):
        raise NotImplementedError

    @abc.abstractmethod
    def find(self, collection, query=None):
        raise NotImplementedError

    @abc.abstractmethod
    def insert(self, collection, records, return_result=False):
        raise NotImplementedError

    @abc.abstractmethod
    def drop(self, collection, query, return_result=False):
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, collection, query, new_vals, return_result=False):
        raise NotImplementedError


class MongoRepository(AbstractRepository):
    def __init__(self, client, db):
        self.client = client
        self.db = self.client[db]

    def create_collection(self, collection, overwrite=True):
        if collection in self.db.list_collection_names():
            if not overwrite:
                raise Exception(
                    f"Collection {collection} already exists. Choose another name, or set `overwrite` to `True`."
                )
            else:
                self.db[collection].drop()

        self.db.create_collection(collection)

    def drop_collection(self, collection):
        if collection in self.db.list_collection_names():
            self.db[collection].drop()

    def find(self, collection, query=None):
        if query is None:
            q = {}
        else:
            q = query
        return self.db[collection].find(q)

    def insert(self, collection, records, return_result=False):
        r = self.db[collection].insert_many(records)
        if return_result:
            return {
                "acknowledged": r.acknowledged,
                "inserted_ids": r.inserted_ids,
            }

    def drop(self, collection, query, return_result=False):
        r = self.db[collection].delete_many(query)
        if return_result:
            return {
                "acknowledged": r.acknowledged,
                "deleted_count": r.deleted_count,
            }

    def update(self, collection, query, new_vals, return_result=False):
        nv = {"$set": {new_vals}}
        r = self.db[collection].update_many(query, nv)
        if return_result:
            return {
                "acknowledged": r.acknowledged,
                "matched_count": r.matched_count,
                "modified_count": r.modified_count,
            }
